from vcScript import *
from vcHelpers.Selection import *
import vcVector

#11.04.24 - Added fucntionality for resetting trigger enabled signal

status = "Grasping"
hold_until = None
finalized = False
trigger_joint = None
grasping_signal = None
trigger_node = None
default_detection_volume = vcVector.new(5,5,5)

def product_transition(product,arrived):
  global status, hold_until

  if arrived == True: # GRASP
    status = "HoldingUntilMinTimeDone" 
  else: # RELEASE
    status = "Grasping" 
    hold_until = None
  evaluateCondition()


#Register to the OnNodeTransform event
def OnStart():
  global trigger_joint, trigger_node, status
  if control_mode_prop is not None and control_mode_prop.Value != "Emulation":
   return
  trigger_node = get_trigger_node(trigger_joint_prop)
  if trigger_node is None:
    print comp.Name, "Servo trigger failed to initialize trigger node."
    return
  #if grasp_mode_p.Value == "DetectionVolume":
  trigger_joint = trigger_node.Dof
  trigger_node.OnNodeTransform = release_trigger_handler
  status = "Grasping"
  
  
  if grasp_mode_p.Value == "ServoTrigger":
    trigger_node.OnNodeTransform = servo_trigger_handler #TODO combine handlers

#Executed after OnStart when simualtion is started
def OnRun():
  global status
  if control_mode_prop is not None and control_mode_prop.Value != "Emulation":
    return
  if collision_trigger_enabled_prop.Value: #Set the collision grasping signal on by default when simulation starts
    collision_grasping_enabled_signal.signal(True)
  
  #TODO Make sure that grasp and release trigger modes control the onrun logic properly
  if grasp_mode_p.Value != "DetectionVolume": #This only applies with collision detection mode? Not really
    return

  while True:
    # condition(lambda: status == "Grasping" or status == "HoldingUntilMinTimeDone")
    condition(lambda: status != "ReadyToRelease" and collision_grasping_enabled_signal.Value)
    if status == "HoldingUntilMinTimeDone":
      delay(delay_grasp.Value)
      condition(lambda: status == "HoldingUntilMinTimeDone")
      status = "ReadyToRelease"
    elif status == "Grasping":
      grasping_signal.signal(True)
      delay(interval_p.Value)
      if container.Components and release_mode_p.Value == "ReleaseAfterDelay":
        delay(release_after_delay_prop.Value)
        grasping_signal.signal(False)
        status = "WaitUntilRegraspDelayDone"
    elif status == "WaitUntilRegraspDelayDone":
      delay(regrasp_delay_prop.Value)
      status = "Grasping"


#Unregister to the OnNodeTransform event
def OnReset():
  global trigger_node
  if trigger_node is not None:
    trigger_node.OnNodeTransform = None

#return the node that should trigger the grasping
def get_trigger_node(trigger_prop):
  if trigger_prop.Value is None:
    return None
  elif trigger_prop.Value == "Tool external parent joint":
    tool_ife = find_tool_interface()
    if tool_ife is None:
      print comp.Name, "No valid interfaces found for external parent joint."
      return None
    trigger_node = find_tool_parent_link_node(tool_ife)
    if trigger_node is None:
      print comp.Name, "Tool interface is not connected."
      return None
    if trigger_node.Name == "mountplate":
      trigger_node = trigger_node.Parent
    return trigger_node
  split_name = trigger_prop.Value.split(":")
  split_name = split_name[0].strip()
  trigger_node = comp.findNode(split_name)
  return trigger_node

#If this component is a robot tool then this will return the interface used to connect to a robot
def find_tool_interface():
  EOAT_ife = comp.findBehaviour("EOAT_MountInterface")
  if EOAT_ife is not None:
    return EOAT_ife
  one_to_one_ifes = comp.findBehavioursByType(VC_ONETOONEINTERFACE)
  connected_ifes = [x for x in one_to_one_ifes if x.IsConnected and not x.IsAbstract]
  if len(connected_ifes) == 0:
    return None
  elif len(connected_ifes) == 1:
    return connected_ifes[0]
  #Return first match. This should be improved if needed
  return connected_ifes[0]

#Try to find a connected section and the parent node of that connected section (from the parent component)
def find_tool_parent_link_node(tool_ife):
  for section in tool_ife.Sections:
    other_section = section.ConnectedToSection
    if other_section is None:
      continue
    other_parent_node = other_section.Interface.Parent
    if other_parent_node is None:
      continue
    return other_parent_node
  return None


#Main event handler for triggering the grasp signal
def release_trigger_handler(arg):
  global status, last_peak_value, last_value, last_direction
  if not sim.IsRunning:
    return
  
  delta = trigger_delta_prop.Value

  #
  if release_mode_p.Value in ("Increasing to Decreasing", "Increasing to Decreasing with TriggerPoint"):
    if trigger_joint.VALUE < last_value: # value decreasing
      if last_direction == "increasing" and abs(trigger_joint.VALUE-last_peak_value) > delta:
        last_direction = "decreasing"
        if status == "ReadyToRelease":
          grasping_signal.signal(False)
          status = "WaitUntilRegraspDelayDone"
          evaluateCondition()
      else:
        last_direction = "increasing"
        last_peak_value = trigger_joint.VALUE
    elif release_mode_p.Value == "Increasing to Decreasing with TriggerPoint": #Value increasing and trigger point
        evaluate_joint_above_trigger()
  #
  elif release_mode_p.Value in ("Decreasing to Increasing", "Decreasing to Increasing with TriggerPoint"):
    if trigger_joint.VALUE > last_value: # value increasing
      if last_direction == "decreasing" and (last_peak_value-trigger_joint.VALUE) > delta:
        last_direction = "increasing"
        if status == "ReadyToRelease":
          grasping_signal.signal(False)
          status = "WaitUntilRegraspDelayDone"
          evaluateCondition()
      else:
        last_direction = "decreasing"
        last_peak_value = trigger_joint.VALUE
    elif release_mode_p.Value == "Decreasing to Increasing with TriggerPoint": #Value decreasing and trigger point
      evaluate_joint_below_trigger()
  #
  elif release_mode_p.Value == "Over TriggerPoint":
    evaluate_joint_above_trigger()
  #
  elif release_mode_p.Value == "Under TriggerPoint":
    evaluate_joint_below_trigger()

  last_value = trigger_joint.VALUE


#Check if joint is below the trigger point
def evaluate_joint_below_trigger():
  global status
  if trigger_joint.VALUE <= trigger_point_prop.Value:
    if status == "ReadyToRelease":
      grasping_signal.signal(False)
      status = "WaitUntilRegraspDelayDone"
      evaluateCondition()

#Check if joint is above the trigger point
def evaluate_joint_above_trigger():
  global status
  if trigger_joint.VALUE >= trigger_point_prop.Value:
    if status == "ReadyToRelease":
      grasping_signal.signal(False)
      status = "WaitUntilRegraspDelayDone"
      evaluateCondition()



#Main event handler for triggering the grasp signal with ServoTrigger
def servo_trigger_handler(arg):
  if not sim.IsRunning:
    return
  if is_closed(trigger_joint):
    if grasping_signal.Value != True:
      grasping_signal.signal(True)
  elif is_open(trigger_joint):
    if grasping_signal.Value != False:
      grasping_signal.signal(False)

#Return difference between current joint value and the trigger point
def get_delta(joint):
  return joint.VALUE - trigger_point_prop.Value

def is_closed(joint):
  #return abs(abs(finger_joint.VALUE) - CLOSED_VAL) < EPSILON
  if invert_prop.Value:
    return get_delta(joint) < trigger_delta_prop.Value
  return get_delta(joint) > (-trigger_delta_prop.Value)

def is_open(joint):
  #return abs(abs(finger_joint.VALUE) - OPEN_VAL) < EPSILON
  if invert_prop.Value:
    return get_delta(joint) > (-trigger_delta_prop.Value)
  return get_delta(joint) < trigger_delta_prop.Value


#Update the trigger joint property step values when the comp nodes are changed (new node might have been added)
def update_trigger_options(type_constant, node1, node2):
  if sim.IsRunning: #Don't change anything when the simulation is running
    return
  if type_constant == VC_NODE_REMOVE_NODE and node1 == comp: #Component deleted
    return
  elif type_constant != VC_NODE_REMOVE_NODE: #Nodes were modified
    set_trigger_servo(node1)
  else: #Node other than the component was removed
    set_trigger_servo(comp)

#Executed when the component is loaded
def OnFinalize():
  global grasping_signal, finalized, container, collision_grasping_enabled_signal
  if finalized:
    return
  finalized = True
  grasping_signal = grasping_node.getBehaviour(grasping_node.Name + " Signal")
  collision_grasping_enabled_signal = grasping_node.getBehaviour(grasping_node.Name + " CollisionTriggerEnabled")
  container = grasping_node.getBehaviour(grasping_node.Name + " Container__HIDE__")
  if grasping_signal is None:
    print "Error. Failed to find the grasping signal"
  container.OnTransition = product_transition
  init_trigger_joint()
  #toggle_hide_trigger_props(servo_trigger_enabled_prop)

#Initializes the TriggerJoint value and step constraints
def init_trigger_joint():
  parent_node = grasping_node.Parent
  trigger_node = set_trigger_servo(parent_node)
  if trigger_node is None:
    print "Failed to set the TriggerJoint"

#Initialize Trigger Servo to the other child of the grasping node
def set_trigger_servo(selected_node):
  children = selected_node.Children
  all_children = getGivenComponentsNodes(comp)
  #Combine the two lists without duplicates
  children = list(children)
  children.extend(x for x in all_children if x not in children)
  if None in children:
    return None
  nodes_and_dofs, trigger_node = build_trigger_step_values(children)
  if not nodes_and_dofs:
    print("No valid servos found for Trigger Servo")
    return None
  #
  # release_joint_prop.StepValues = nodes_and_dofs
  # if release_joint_prop.Value is None or release_joint_prop.Value == "":
  #   release_joint_prop.Value = release_joint_prop.StepValues[-1]
  
  #
  trigger_joint_prop.StepValues = nodes_and_dofs
  if trigger_joint_prop.Value is None or trigger_joint_prop.Value == "":
    trigger_joint_prop.Value = trigger_joint_prop.StepValues[-1]
  if trigger_node is None:
    trigger_node = trigger_joint_prop.StepValues[-1]

  return trigger_node

#Return a list of Node name & Dof name, and the first trigger node to be the default value
def build_trigger_step_values(children):
  trigger_node = None
  nodes_and_dofs = []
  VALID_JOINT_TYPES = [VC_JOINTTYPE_TRANSLATIONAL, VC_JOINTTYPE_ROTATIONAL, VC_JOINTTYPE_CUSTOM]
  for c in children:
    if c.Dof is None or c.Dof.JointLinkType not in VALID_JOINT_TYPES:
      continue
    if trigger_node == None:
      trigger_node = c
    combined_names = "{} : {}".format(c.Name, c.Dof.Name)
    nodes_and_dofs.append(combined_names)
  nodes_and_dofs.insert(0, "Tool external parent joint")
  return nodes_and_dofs, trigger_node

def force_prod_locs(p):
  if force_locs_p.Value:
    force_pick_p.IsVisible = True
    force_place_p.IsVisible = True
    print "Type in the axis that you want to force in ForceProductOffset: X, Y, and/or Z"
  else:
    force_pick_p.IsVisible = False
    force_place_p.IsVisible = False

#Validate that the given property exists, and return the property
def confirmProperty(name,defaultvalue,type=VC_REAL,constraints=VC_PROPERTY_DEFAULT, quantity=None):
  prop = comp.getProperty(name)
  if type == VC_VECTOR:
    if prop is None:
      prop = comp.createProperty(type,name,constraints)
      if defaultvalue:
        prop.Value = vcVector.new()
      else:
        prop.Value = default_detection_volume
  else:
    if prop is None:
      if constraints is None:
        prop = comp.createProperty(type,name)
      else:
        prop = comp.createProperty(type,name,constraints)
      if constraints == VC_PROPERTY_STEP:
        prop.StepValues = DEFAULT_TRIGGER_JOINT_VALS
      prop.Value = defaultvalue
  if type == VC_REAL and quantity is not None:
    prop.Quantity = quantity
  return prop

#Use the property to set the signal value if needed
def set_collison_trigger_state(arg):
  if sim.IsRunning:
    collision_grasping_enabled_signal.signal(arg.Value)

# #hide/show the trigger properties if the trigger is enabled or not
# def toggle_hide_trigger_props(arg):
#   #trigger_joint_prop.IsVisible = arg.Value
#   #trigger_point_prop.IsVisible = arg.Value
#   trigger_delta_prop.IsVisible = arg.Value
#   invert_prop.IsVisible = arg.Value


comp = getComponent()
sim = getSimulation()
grasping_node = getNode()


DEFAULT_EPSILON = 0.5
DEFAULT_TRIGGER_POINT = 5.0
DEFAULT_TRIGGER_JOINT = ''
DEFAULT_INTERVAL = 0.05
DEFAULT_MIN_PRODUCT_HOLD = 0.1 
DEFAULT_MIN_REGRASP_DELAY = 0.3
DEFAULT_RELEASE_TRIGGER_JOINT = '' 
DEFAULT_RELEASE_TRIGGER_MODE = 'Signal'
DEFAULT_GRASP_MODE = "Signal"
DEFAULT_TRIGGER_JOINT_VALS = []
DEFAULT_FORCE_PICK_PLACE_LOCATIONS = ""
DEFAULT_FORCE_PICK = 0.0
DEFAULT_FORCE_PLACE = 0.0
DEFAULT_COLLISION_TRIGGER_MODE = False
DEFAULT_RELEASE_DELAY = 0.3

# DEFAULT_RELEASE_TRIGGER_MODE_VALS = []

#Servo trigger properties
#servo_trigger_enabled_prop = confirmProperty(grasping_node.Name + "::ServoTriggerEnabled", False, VC_BOOLEAN)
trigger_label_prop = comp.getProperty(grasping_node.Name + "::TriggerSettings")
if trigger_label_prop is None:
  trigger_label_prop = comp.createProperty(VC_BUTTON, grasping_node.Name + "::TriggerSettings")
  trigger_label_prop.WritableWhenConnected = False
  trigger_label_prop.WritableWhenDisconnected = False
  trigger_label_prop.WritableWhenSimulating = False

grasp_mode_p = confirmProperty(grasping_node.Name + "::GraspTriggerMode", DEFAULT_RELEASE_TRIGGER_MODE, VC_STRING, VC_PROPERTY_STEP)
grasp_mode_p.StepValues = ["Signal", "ServoTrigger", "DetectionVolume"]
if grasp_mode_p.Value not in grasp_mode_p.StepValues:
  grasp_mode_p.Value = grasp_mode_p.StepValues[0]
release_mode_p = confirmProperty(grasping_node.Name + "::ReleaseTriggerMode", DEFAULT_RELEASE_TRIGGER_MODE, VC_STRING, VC_PROPERTY_STEP)
release_mode_p.StepValues = ["Signal", "Decreasing to Increasing","Increasing to Decreasing", "Decreasing to Increasing with TriggerPoint","Increasing to Decreasing with TriggerPoint", "Over TriggerPoint", "Under TriggerPoint", "ReleaseAfterDelay"]
#
trigger_joint_prop = confirmProperty(grasping_node.Name + "::TriggerJoint", DEFAULT_TRIGGER_JOINT, VC_STRING, VC_PROPERTY_STEP)
trigger_point_prop = confirmProperty(grasping_node.Name + "::TriggerPoint", DEFAULT_TRIGGER_POINT, VC_REAL)
trigger_delta_prop = confirmProperty(grasping_node.Name + "::TriggerDelta", DEFAULT_EPSILON, VC_REAL)
invert_prop = confirmProperty(grasping_node.Name + "::InvertTrigger", False, VC_BOOLEAN)


#Collision trigger properties
last_peak_value = None
last_value = None
last_direction = None
collision_trigger_enabled_prop = confirmProperty(grasping_node.Name + "::CollisionTriggerEnabled", DEFAULT_COLLISION_TRIGGER_MODE, VC_BOOLEAN)
reset_after_grasp_prop = confirmProperty(grasping_node.Name + "::ResetAfterGrasp",False,VC_BOOLEAN)
interval_p = confirmProperty(grasping_node.Name + "::GraspDetectionInterval", DEFAULT_INTERVAL, VC_REAL, None, "Time")
delay_grasp = confirmProperty(grasping_node.Name + "::DelayAfterGrasp", DEFAULT_MIN_PRODUCT_HOLD, VC_REAL, None, "Time")

regrasp_delay_prop = confirmProperty(grasping_node.Name + "::DelayAfterRelease", DEFAULT_MIN_REGRASP_DELAY, VC_REAL, None, "Time")
release_after_delay_prop = confirmProperty(grasping_node.Name +"::ReleaseAfterDelay", DEFAULT_RELEASE_DELAY, VC_REAL, None, "Time")

collision_trigger_enabled_prop.OnChanged = set_collison_trigger_state
force_locs_p = confirmProperty(grasping_node.Name + "::ForceProductOffset", DEFAULT_FORCE_PICK_PLACE_LOCATIONS, VC_STRING)
# force_locs_p.Value = False
force_locs_p.OnChanged = force_prod_locs
force_pick_p = confirmProperty(grasping_node.Name + "::PickOffset", DEFAULT_FORCE_PICK, VC_VECTOR)
force_place_p = confirmProperty(grasping_node.Name + "::PlaceOffset", DEFAULT_FORCE_PLACE, VC_VECTOR)
force_prod_locs(force_locs_p)

# release_joint_p.StepValues = ['Cable : Cable', 'J1 : J1', 'J2 : J2', 'J3 : J3', 'J4 : J4']
control_mode_prop = comp.getProperty("ControlMode")
trigger_point_prop = comp.getProperty(grasping_node.Name +"::TriggerPoint")


det_volume = comp.getProperty(grasping_node.Name + "::DetectionVolumeSize")
det_volume.Value = vcVector.new(5,5,5)
det_volume.IsVisible = True
gravity_dir = comp.getProperty(grasping_node.Name + "::GravityDirection")
gravity_dir.Value = 5

#servo_trigger_enabled_prop.OnChanged = toggle_hide_trigger_props
comp.OnNodeConfigurationChange = update_trigger_options
OnFinalize()


#TODO Toggle which trigger proeprties are visible based on trigger modes